
#include "stdio.h"
#include "stdlib.h"
#include "stdint.h"

#include "gf16.h"

static
int byte_fdump(FILE * fp, const char * extra_msg , const unsigned char *v, unsigned n_byte)
{
    int r = 0;
    int t = fprintf( fp , "const unsigned char %s[%d] __attribute__((aligned(32))) = {\n", extra_msg , n_byte );
    if( 0 > t ) return t;
    else r += t;

    for(unsigned i=0;i<n_byte;i++) {
        int t = fprintf( fp , "0x%02x", v[i] );
        if( 0 > t ) return t;
        else r += t;

        if( i==n_byte-1 ) { r += fprintf( fp , "\n};\n\n"); break; }

        if(i%32 == 31 ) r += fprintf( fp , ",\n");
        else if(i%8 == 7) r += fprintf( fp , ", " );
        else r += fprintf( fp , "," );
    }
    return r;
}



static
void vec_cpy( unsigned char * d , const unsigned char * s, unsigned n ) { for(unsigned i=0;i<n;i++) d[i]=s[i]; }



static
int gen_squ_tab( unsigned char * squ_tab )
{
    for(unsigned i=0;i<16;i++) squ_tab[i] = gf16_mul( i , i );

    vec_cpy( squ_tab+16 , squ_tab , 16 );
    return 32;
}


static
int gen_inv_tab( unsigned char * inv_tab )
{
    for(unsigned i=0;i<16;i++) inv_tab[i] = gf16_inv( i );

    vec_cpy( inv_tab+16 , inv_tab , 16 );
    return 32;
}



static
int gen_exp_tab( unsigned char * exp_tab )
{
    unsigned char x = 2;
    unsigned char a = 1;

    exp_tab[0] = a;
    for(int i=1;i<16;i++) {
        a = gf16_mul( a , x );
        exp_tab[i] = a;
    }

    vec_cpy( exp_tab + 16 , exp_tab , 16 );
    return 32;
}


static
unsigned find( const unsigned char * vec , unsigned char tar , unsigned len ) { for(unsigned i=0;i<len;i++) { if(vec[i]==tar) return i; } return -1; }

static
int gen_log_tab( unsigned char * log_tab , const unsigned char * exp_tab )
{
    log_tab[0] = -42;
    for(unsigned i=1;i<16;i++) log_tab[i] = find( exp_tab , i , 16 );

    vec_cpy( log_tab + 16 , log_tab , 16 );
    return 32;
}





static
int gen_gf16_multab( unsigned char * multab )
{
    for(int i=0;i<16;i++) {
        for(int j=0;j<16;j++) multab[i*32+j] = gf16_mul( i , j );
        vec_cpy( multab+i*32+16 , multab+i*32 , 16 );
    }

    return 16*2*16;
}


static
int gen_gf16_multab_base( unsigned char *mulbase )
{

    for(int i=0;i<4;i++) {
        for(int j=0;j<16;j++) mulbase[i*32+j] = gf16_mul( 1<<i , j );
        vec_cpy( mulbase+i*32+16 , mulbase+i*32 , 16 );
    }

    return 16*2*4;

}




//////////////////////////////////



static
int gen_gf256_multab( unsigned char * multab )
{
    for(int i=0;i<256;i++) {
        for(int j=0;j<16;j++) multab[i*32+j]    = gf256_mul( i , j );
        for(int j=0;j<16;j++) multab[i*32+j+16] = gf256_mul( i , j<<4 );
    }

    return 16*2*256;
}


static
int gen_gf256_multab_base( unsigned char *multab )
{

    for(int i=0;i<8;i++) {
        for(int j=0;j<16;j++) multab[i*32+j]    = gf256_mul( 1<<i , j );
        for(int j=0;j<16;j++) multab[i*32+j+16] = gf256_mul( 1<<i , j<<4 );
    }

    return 16*2*8;
}




static
int gen_gf256_squtab( unsigned char *squtab )
{

    for(int j=0;j<16;j++) squtab[j]    = gf256_mul( j , j );
    for(int j=0;j<16;j++) squtab[j+16] = gf256_mul( j<<4 , j<<4 );

    return 16*2;
}







int main()
{

    fprintf( stdout , "// Contents are generated by %s.\n\n" , __FILE__ );
    fprintf( stdout , "#include \"gf16_tabs.h\"\n\n" );

    unsigned char squ_tab[32];
    gen_squ_tab( squ_tab );
    byte_fdump( stdout , "__gf16_squ" , squ_tab , 32 );

    unsigned char inv_tab[32];
    gen_inv_tab( inv_tab );
    byte_fdump( stdout , "__gf16_inv" , inv_tab , 32 );

    unsigned char exp_tab[32];
    gen_exp_tab( exp_tab );
    byte_fdump( stdout , "__gf16_exp" , exp_tab , 32 );


    unsigned char log_tab[32];
    gen_log_tab( log_tab , exp_tab );
    byte_fdump( stdout , "__gf16_log" , log_tab , 32 );


    unsigned char gf16_multab[16*2*16];
    gen_gf16_multab( gf16_multab );
    byte_fdump( stdout , "__gf16_mul" , gf16_multab , sizeof(gf16_multab) );


    unsigned char gf16_mulbase[16*2*4];
    gen_gf16_multab_base( gf16_mulbase );
    byte_fdump( stdout , "__gf16_mulbase" , gf16_mulbase , sizeof(gf16_mulbase) );


//////////////////////////////////

    unsigned char gf256_squtab[16*2];
    gen_gf256_squtab( gf256_squtab );
    byte_fdump( stdout , "__gf256_squ" , gf256_squtab , sizeof(gf256_squtab) );

    unsigned char gf256_mulbase[16*2*8];
    gen_gf256_multab_base( gf256_mulbase );
    byte_fdump( stdout , "__gf256_mulbase" , gf256_mulbase , sizeof(gf256_mulbase) );

    unsigned char gf256_multab[16*2*256];
    gen_gf256_multab( gf256_multab );
    byte_fdump( stdout , "__gf256_mul" , gf256_multab , sizeof(gf256_multab) );





    return 0;
}

